from unittest import TestCase

import os
import time
import shutil


class BaseTestCase(TestCase):
    # In case of assert failures, print the full error message
    maxDiff = None

    @staticmethod
    def get_sample_path(filename=""):
        samples_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "samples")
        _, ext = os.path.splitext(filename)
        if ext == '':
            filename += ".yml"
        return os.path.join(samples_dir, filename)

    @staticmethod
    def get_expected(filename=""):
        expected_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "expected")
        _, ext = os.path.splitext(filename)
        if ext == '':
            filename += ".yml"
        expected_path = os.path.join(expected_dir, filename)
        expected = open(expected_path).read()
        return expected

    @staticmethod
    def copy_sample(samplename):
        return TempSampleCopy(samplename)


class TempSampleCopy(object):
    """ Simple context manager that copies a given sample file and makes it path available in it's context.
        After exiting the context, the sample file copy is deleted."""

    def __init__(self, samplename):
        self.sample_path = BaseTestCase.get_sample_path(samplename)
        filename = os.path.basename(self.sample_path)
        copy_filename = "copy_%s_%s" % (filename, time.strftime("%Y-%m-%d-%H-%M-%S"))
        sample_dir = os.path.dirname(self.sample_path)
        self.copy_path = os.path.join(sample_dir, copy_filename)

    def __enter__(self):
        shutil.copyfile(self.sample_path, self.copy_path)
        return self.copy_path

    def __exit__(self, type, value, tb):
        os.remove(self.copy_path)



from yamlpal.tests.base import BaseTestCase
from yamlpal import cli
from click.testing import CliRunner


class FindNegativeTests(BaseTestCase):
    def setUp(self):
        self.cli = CliRunner()

    def test_inaccurate_path_dictionary(self):
        result = self.cli.invoke(cli.cli, ["find", "bill-to/address", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, "ERROR: Path exists but not specific enough (bill-to/address).\n")

    def test_inaccurate_path_list(self):
        result = self.cli.invoke(cli.cli, ["find", "product", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, "ERROR: Path exists but not specific enough (product).\n")



from yamlpal.tests.base import BaseTestCase
from yamlpal import cli
from click.testing import CliRunner


class FindTests(BaseTestCase):
    def setUp(self):
        self.cli = CliRunner()

    def test_find_simple(self):
        result = self.cli.invoke(cli.cli, ["find", "title", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, "title: Invoice for purchases\n")

    def test_find_number(self):
        result = self.cli.invoke(cli.cli, ["find", "tax", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, "tax: 251.42\n")

    def test_find_complex(self):
        result = self.cli.invoke(cli.cli, ["find", "bill-to/address/city", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, "city: Royal Oak\n")

    def test_find_custom_format(self):
        sample_path = self.get_sample_path("sample1")
        result = self.cli.invoke(cli.cli, ["find", "bill-to/address/city",
                                           "-F", "%{file} %{linenr} %{key} %{value}",
                                           "-f", sample_path])
        self.assertEqual(result.output, "%s 11 city Royal Oak\n" % sample_path)

    def test_find_custom_format_literal(self):
        result = self.cli.invoke(cli.cli, ["find", "bill-to/address/city",
                                           "-F", "%{literal}",
                                           "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, "        city    : Royal Oak\n")

    def test_find_key_of_list_item(self):
        result = self.cli.invoke(cli.cli, ["find", "labels[0]",
                                           "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, "0: premium customer\n")

        result = self.cli.invoke(cli.cli, ["find", "labels[1]",
                                           "-f", self.get_sample_path("sample1")])

        self.assertEqual(result.output, "1: online order\n")



from yamlpal.tests.base import BaseTestCase
from yamlpal import cli
from click.testing import CliRunner


class InsertionTests(BaseTestCase):
    def setUp(self):
        self.cli = CliRunner()

    def test_insert_inline(self):
        with self.copy_sample("sample1") as copy_path:
            result = self.cli.invoke(cli.cli, ["insert", "--inline", "title", "newkey: value", "-f", copy_path])
            actual = open(copy_path).read()
            self.assertEqual(result.output, "")
            self.assertEqual(actual, self.get_expected("sample1-after-string"))

    def test_insert_ignore_leading_trailing_whitespace(self):
        result = self.cli.invoke(cli.cli, ["insert", "title", "   newkey: value \t \n",
                                           "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, self.get_expected("sample1-after-string"))

    def test_insert_after_date(self):
        result = self.cli.invoke(cli.cli, ["insert", "date", "newkey: value", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, self.get_expected("sample1-after-date"))

    def test_insert_after_int(self):
        result = self.cli.invoke(cli.cli, ["insert", "invoice", "newkey: value", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, self.get_expected("sample1-after-int"))

    def test_insert_after_float(self):
        result = self.cli.invoke(cli.cli, ["insert", "tax", "newkey: value", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, self.get_expected("sample1-after-float"))

    def test_insert_after_string(self):
        result = self.cli.invoke(cli.cli, ["insert", "title", "newkey: value", "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, self.get_expected("sample1-after-string"))

    def test_multiple_file_flags(self):
        # default case: multiple file flags
        result = self.cli.invoke(cli.cli, ["insert", "title", "newkey: value",
                                           "-f", self.get_sample_path("sample1"),
                                           "-f", self.get_sample_path("sample2")])
        output1 = self.get_expected("sample1-after-string")
        output2 = self.get_expected("sample2-after-string")
        self.assertEqual(result.output, output1 + output2)

        # default case: swapped order -> yamlpal maintains the order
        result = self.cli.invoke(cli.cli, ["insert", "title", "newkey: value",
                                           "-f", self.get_sample_path("sample2"),
                                           "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, output2 + output1)

        # Same file 3 times (yamlpal does NOT select unique filepaths, it does exactly what you tell it to do)
        result = self.cli.invoke(cli.cli, ["insert", "title", "newkey: value",
                                           "-f", self.get_sample_path("sample1"),
                                           "-f", self.get_sample_path("sample1"),
                                           "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, output1 + output1 + output1)

    def test_multiple_stdin_files(self):
        # default case:  multiple files via stdin
        stdin = "%s\n%s" % (self.get_sample_path("sample1"), self.get_sample_path("sample2"))
        output1 = self.get_expected("sample1-after-string")
        output2 = self.get_expected("sample2-after-string")
        result = self.cli.invoke(cli.cli, ["insert", "title", "newkey: value"], input=stdin)
        self.assertEqual(result.output, output1 + output2)

        # default case: swapped order -> yamlpal maintains the order
        stdin = "%s\n%s" % (self.get_sample_path("sample2"), self.get_sample_path("sample1"))
        result = self.cli.invoke(cli.cli, ["insert", "title", "newkey: value"], input=stdin)
        self.assertEqual(result.output, output2 + output1)

        # combination of stdin and file flags
        stdin = "%s\n%s" % (self.get_sample_path("sample1"), self.get_sample_path("sample2"))
        output3 = self.get_expected("sample3-after-string")
        result = self.cli.invoke(cli.cli, ["insert", "title", "newkey: value",
                                           "-f", self.get_sample_path("sample1"),
                                           "-f", self.get_sample_path("sample3")],
                                 input=stdin)
        self.assertEqual(result.output, output1 + output3 + output1 + output2)

    def test_insert_content_from_file(self):
        result = self.cli.invoke(cli.cli, ["insert", "title", "@" + self.get_sample_path("insert-multiline.txt"),
                                           "-f", self.get_sample_path("sample1")])
        self.assertEqual(result.output, self.get_expected("sample1-multiline-insert"))



import yamlpal
from yamlpal.yaml_parser import YamlParser
from yamlpal import exceptions
import sys
import click
import re
import os

FORMATTING_EPILOG = """
\b
Format Strings:
Format strings determine how yamlpal will output values.
Available keys:
   %{key}     :  Key of the match (or index if matching an item in a list)
   %{value}   :  Value of the match
   %{linenr}  :  Line number where the match occured
   %{file}    :  Name of the file in which the match occured
   %{literal} :  Literal match in the file (original formatting)
Examples:
   $ yamlpal find "bill-to/address/city" --format "%{file} %{linenr}: %{value}"
   /abs/path/to/examples/examples/sample1.yml 11: Royal Oak

   $ yamlpal find "bill-to/address/city" --format "%{linenr} %{literal}"
   11:         city    : Royal Oak
"""


@click.group(epilog="Run 'yamlpal <command> --help' for command specific help.")
@click.version_option(version=yamlpal.__version__)
def cli():
    """ Modify and search yaml files while keeping the original formatting.
    """


def get_files(passed_files):
    """ Determines which files are part of the list that will be manipulated by yamlpal by combining
    the files that are passed as commandline arguments with the files that are passed via the stdin.
    :param passed_files: list of files that is passed via cli flags (-f).
    :return: list of files that will be manipulated by yamlpal
    """
    all_files = []
    all_files.extend(passed_files)
    if not sys.stdin.isatty():
        input_paths = sys.stdin.read().split("\n")
        for input_path in input_paths:
            if input_path.strip() != "":
                all_files.append(os.path.abspath(input_path))
                # TODO(jroovers): check if valid file paths
    return all_files


def get_str_content(str_value):
    """ Returns the string content of a passed yaml content value (for an insert/replace yamlpal operation).
        If the passed str_value starts with an '@' then we attempt to treat the passed string as a filename and
        read the contents of the file."""
    if str_value.startswith("@"):
        file_path = str_value[1:]

        if not os.path.isfile(file_path):
            click.echo("ERROR: Invalid file content path '%s'." % file_path, err=True)
            exit(1)

        with open(file_path, 'r') as f:
            content = f.read()
            # strip off newline at the end if it's there: insert/replace takes care of this
            if content.endswith("\n"):
                content = content[0:-1]
    else:
        # If we directly pass the string, strip whitespace and allow newline and tab chars
        content = str_value.strip().replace("\\n", "\n").replace("\\t", "\t")
    return content


@cli.command("insert")
@click.argument('needle')
@click.argument('newcontent')
@click.option('-f', '--file', type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
              multiple=True, help="File to insert new content in. Can by specified multiple times to modify " +
                                  "multiple files. Files are not modified inline by default. " +
                                  "You can also provide (additional) file paths via stdin.")
@click.option('-i', '--inline', help="Edit file inline instead of dumping it to std out.", is_flag=True)
def insert(needle, newcontent, file, inline):
    """ Insert new content into a yaml file. """
    newcontent = get_str_content(newcontent)

    files = get_files(file)
    for file in files:
        insert_in_file(needle, newcontent, file, inline)


@cli.command("find", epilog=FORMATTING_EPILOG)
@click.argument('needle')
@click.option('-f', '--file', type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
              help="File to find content in.")
@click.option('-F', '--format', help="Format string in which matched content should be returned. " +
                                     "See the section 'Format Strings' below for details on format strings. " +
                                     "(default: \"%{key}: %{value}\")",
              default="%{key}: %{value}")
def find(needle, file, format):
    """ Find a line in a yaml file. """
    found = find_in_file(needle, file, format)
    click.echo(found)


def find_in_file(needle, file, format):
    # read yaml file
    fp = open(file)
    filecontents = fp.read()
    fp.close()

    # parse the file
    data = YamlParser.load_yaml(filecontents)
    try:
        element = find_element(data, needle)
    except exceptions.InvalidSearchStringException:
        # TODO (jroovers): we should deduplicate this code. Best done by moving the core business logic
        # (like find_element) out of this module into it's own module and then creating a wrapper function
        # here that deals with exception handling
        click.echo("ERROR: Invalid search string '%s' for file '%s'" % (needle, file), err=True)
        exit(1)

    return apply_format(file, filecontents, element, format)


def apply_format(file, filecontents, element, format):
    """ Given a yaml element and yamlpal format string, return the interpolated string.
        We currently support the following placeholders:
         - %{key}     -> key of the yaml element (index if you are accessing a list)
         - %{value}   -> value of the yaml element
         - %{literal} -> the string corresponding to the yaml element as it literally occurs in the file
         - %{linenr}  -> line number on which the yaml element is found
         - %{file}    -> name of the file in which the yaml element is found
    """

    result = format.replace("%{key}", str(element.key))
    result = result.replace("%{value}", str(element))
    result = result.replace("%{linenr}", str(element.line))
    result = result.replace("%{file}", file)

    # check whether literal occurs before splitting the file, since it's a more expensive operation
    if "%{literal}" in format:
        lines = filecontents.split("\n", element.line + 1)  # don't split more than required
        result = result.replace("%{literal}", lines[element.line])

    return result


def insert_in_file(needle, newcontent, file, inline):
    # read yaml file
    fp = open(file)
    filecontents = fp.read()
    fp.close()

    # parse yaml, find target line, inject new line
    data = YamlParser.load_yaml(filecontents)
    try:
        element = find_element(data, needle)
    except exceptions.InvalidSearchStringException:
        click.echo("ERROR: Invalid search string '%s' for file '%s'" % (needle, file), err=True)
        exit(1)

    updated_filecontents = insert_line(element.line, newcontent, filecontents)

    # write new content to file or stdout
    if inline:
        fp = open(file, "w")
        fp.write(updated_filecontents)
        fp.close()
    else:
        click.echo(updated_filecontents, nl=False)


def find_element(yaml_dict, search_str):
    """ Given a dictionary representing a yaml document and a yaml path string, find the specified element in the
        dictionary."""

    # First split on / to determine which yaml dict we are searching in
    dict_parts = search_str.split("/")
    parsed_parts = []

    for dict_part in dict_parts:
        matches = re.match("(.*)(\[([0-9]+)\])", dict_part)
        if matches:
            list_name = matches.groups()[0]
            list_index = int(matches.groups()[2])
            parsed_parts.append(list_name)
            parsed_parts.append(list_index)
        else:
            parsed_parts.append(dict_part)

    # traverse the yaml path
    node = yaml_dict
    try:
        for key in parsed_parts:
            node = node[key]
    except (KeyError, IndexError):
        raise exceptions.InvalidSearchStringException(search_str)

    # Try accessing the line of the path we are currently on. If we can't access it,
    # it means that the user has specified a path to a dict or list, without indicating an item within the
    # dictionary or list.
    try:
        node.line
        node.key = parsed_parts[-1]  # add the last parsed key as the node's key
    except AttributeError:
        click.echo("ERROR: Path exists but not specific enough (%s)." % search_str, err=True)
        exit(1)
    return node


def insert_line(line_nr, new_content, filecontents):
    lines = filecontents.split("\n")

    # determine the size of indentation of the line we searched for so that we can use the same indentation
    indentation_size = len(lines[line_nr]) - len(lines[line_nr].lstrip())
    # copy indentation so we use the same whitespace characters (tab, space, mix of tab and space)
    indentation_chars = lines[line_nr][0:indentation_size]

    new_content = indentation_chars + new_content
    lines.insert(line_nr + 1, new_content)

    newfile = "\n".join(lines)
    return newfile


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
from setuptools import setup, find_packages
import re
import os

# There is an issue with building python packages in a shared vagrant directory because of how setuptools works
# in python < 2.7.9. We solve this by deleting the filesystem hardlinking capability during build.
# See: http://stackoverflow.com/a/22147112/381010
del os.link

description = "Simple tool for inserting new entries in yaml files while keeping the original structure and formatting"
long_description = """
Simple tool for inserting new entries in yaml files while keeping the original structure and formatting.

Source code on `github.com/jorisroovers/yamlpal`_.

.. _github.com/jorisroovers/yamlpal: https://github.com/jorisroovers/yamlpal
"""


# shamelessly stolen from mkdocs' setup.py: https://github.com/mkdocs/mkdocs/blob/master/setup.py
def get_version(package):
    """Return package version as listed in `__version__` in `init.py`."""
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


setup(
    name="yamlpal",
    version=get_version("yamlpal"),
    description=description,
    long_description=long_description,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License"
    ],
    install_requires=[
        'Click==6.2',
        'PyYAML==3.11'
    ],
    keywords='yamlpal yaml',
    author='Joris Roovers',
    url='https://github.com/jorisroovers/yamlpal',
    license='MIT',
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "yamlpal = yamlpal.cli:cli",
        ],
    },
)



import yaml
from yaml.composer import Composer
from yaml.constructor import Constructor


class LineStr(str):
    """ String that has an associated line number """
    pass


class YamlParser(object):
    """ Yaml parser that attaches line numbers to a dictionary that is parsed by pyYAML.

    This parser monkeypatches some internal PyYAML methods in order to add line numbers to the returned dictionary.

    In particular:

        1. We decorate PyYAML's compose_node(...) method to add line numbers to PyYAML's internal yaml node
           representation at the time tokens are read from the yaml file.

        2. When nodes are constructed into python objects, we retrieve the line number from the node representation
           and store these in a LineStr object. LineStr objects (see class defined above) are just strings that have
           an associated line number.


    Based on http://stackoverflow.com/questions/13319067/parsing-yaml-return-with-line-number
    """

    @staticmethod
    def load_yaml(filecontents):

        loader = yaml.Loader(filecontents)

        def compose_node(parent, index):
            """ Invoked when a new node (key, value or compound (dict, list) type) is created. """
            line = loader.line  # the line number where the previous token has ended (plus empty lines)

            # call the original compose_node
            node = Composer.compose_node(loader, parent, index)

            # In case we are dealing with a yaml list, the line numbers turn out to be off by 1 which we need to fix.
            # What the parser will do is first invoke this method with every item in the list and then invoke it again
            # with the entire list. At that point we can iterate over the list and substract 1 from the line numbers
            # of the items in the list.
            # For single items in a list, Composer.compose_node(...) returns a ScalarNode. node.value is then of type
            # str, int, float, etc. For the entire list, Composer.compose_node(...) will return a SequenceNode with
            # node.value being of type 'list'. So by checking for the type of node.value, we can figure out whether
            # we are currently processing an item in a list or the actual list itself.
            # The only additional required check is making sure that in case node.value is a list, the items in
            # node.value are not tuples (because if they are, we are dealing with a nested dictionary instead of a
            # normal list). We can do this by checking whether node.value[0] is an instance of tuple or not.
            if isinstance(node.value, list) and len(node.value) > 0:
                if not isinstance(node.value[0], tuple):
                    # We're processing a real yaml list -> substract 1 from the line numbers in the list
                    for list_node in node.value:
                        list_node.value['__line__'] -= 1

            if not isinstance(node.value, dict):
                # replace the value of the node with a dictionary that contains both the line number and value
                node.value = {"__line__": line, "__val__": node.value}

            return node

        def construct_object(node, deep=False):
            """ Invoked when PyYAML's internal nodes are converted to python types
            (e.g. ScalarNode -> str/int/float, SequenceNode -> list, MappingNode -> dictionary)
            """
            # in some not fully understood cases we get passed the strings '__line__' and '__val__'
            # i.e. the extra dict attributes that we created. We're just skipping all nodes that don't
            # have a 'value' attribute here which fixes the problem.
            if not hasattr(node, 'value') or not isinstance(node.value, dict):
                return

            # retrieve previously stored line number and value, restore node.value to original value
            line = node.value['__line__']
            node.value = node.value['__val__']

            # call the original construct_object method
            data = Constructor.construct_object(loader, node, deep)

            # if the original construct_object method creating anything else then a list or dict (i.e. a str, int, float
            # datetime, etc), just wrap it in a LineStr object.
            if not (isinstance(data, dict) or isinstance(data, list)):
                data = LineStr(data)
                data.line = line
            return data

        # monkey patch some methods from the PyYAML loader so that we can work our magic while the yaml
        # mapping is constructed
        loader.compose_node = compose_node
        loader.construct_object = construct_object

        return loader.get_single_data()



from yamlpal.tests.base import BaseTestCase
from yamlpal import cli
from yamlpal import __version__
from click.testing import CliRunner


class CLITests(BaseTestCase):
    def setUp(self):
        self.cli = CliRunner()

    def test_version(self):
        result = self.cli.invoke(cli.cli, ["--version"])
        self.assertEqual(result.output.split("\n")[0], "cli, version {0}".format(__version__))
